/*
 * Copyright (C) 2024 sillysagiri
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <cstdint>
#include <cstdlib>
#include <filesystem>
#include <fstream>
#include <stdexcept>
#include <format>
#include <iostream>
#include <vector>

#define STB_IMAGE_IMPLEMENTATION
#define STB_IMAGE_WRITE_IMPLEMENTATION

#include <argparse/argparse.hpp>
#include "header.hpp"
#include "Timer.h"
#include "fastlz.h"
#include "murmurhash.h"

namespace fs = std::filesystem;

int main(int argc, char **argv)
{
	srand(time(0));
	argparse::ArgumentParser program("sillyimage", "1.3");
	Options opt;

	// Register argument
	{
		program.add_argument("-i", "--input")
			.help("input files")
			.append()
			.required()
			.store_into(opt.path_input);

		program.add_argument("-o", "--out")
			.help("output directory")
			.required()
			.store_into(opt.path_output);

		program.add_argument("-f", "--format")
			.help("texture format { rgb256, rgb16, indexed4, indexed16, indexed256, indexed32a8, indexed8a32 }")
			.default_value<std::string>("rgb16")
			.choices("rgb256", "rgb16", "indexed4", "indexed16", "indexed256", "indexed32a8", "indexed8a32")
			.nargs(1)
			.store_into(opt.format);
	}

	// Parse argument
	{
		try
		{
			program.parse_args(argc, argv);

			// input list
			for (auto &i : opt.path_input)
				if (i.empty() || !fs::exists(i)) 
					throw std::runtime_error(std::format("file doesnt exsist: \"{}\"", i));

			// output directory
			if (opt.path_output.empty() || !fs::is_directory(opt.path_output))
				throw std::runtime_error(std::format("invalid output directory: \"{}\"", opt.path_output));

		}
		catch(const std::exception &e)
		{
			std::cerr << std::format("Error: {}\n\n", e.what()) << program << "\n";
			return 1;
		}
	}

	// Convert
	try {

		if (opt.format == "rgb256")
		{
			for(auto &input : opt.path_input)
			{
				Image img;
				img.Load(input);

				Metadata meta;
				meta.format = Format::Format_RGB_256;
				meta.width = img.width;
				meta.height = img.height;
				meta.palette_count = 0;
				meta.palette_hash = 0;
				meta.original_size = img.data.size();

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(1, img.data.data(), meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(input).stem().string());
				std::ofstream out(output, std::ios::binary);
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}

		if (opt.format == "rgb16")
		{
			for(auto &input : opt.path_input)
			{
				Image img;
				img.Load(input);
				std::vector<uint16_t> img16(img.width*img.height);

				// convert into RGB16 color
				for (int i=0; i<img.width*img.height; i++)
					img16[i] = RGB255_RGB16(img.data[i*4], img.data[i*4+1], img.data[i*4+2], img.data[i*4+3]);

				Metadata meta;
				meta.format = Format::Format_RGB_16;
				meta.width = img.width;
				meta.height = img.height;
				meta.palette_count = 0;
				meta.palette_hash = 0;
				meta.original_size = img16.size()*sizeof(uint16_t);

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(2, img16.data(), meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(input).stem().string());
				std::ofstream out(output, std::ios::binary);
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}

		if (opt.format == "indexed4" ||
				opt.format == "indexed16" ||
				opt.format == "indexed256" ||
				opt.format == "indexed32a8" ||
				opt.format == "indexed8a32")
		{
			uint16_t numcol;
			uint8_t format;

			if (opt.format == "indexed4")
			{
				numcol = 4;
				format = Format_INDEXED_4;
			}
			else if (opt.format == "indexed16")
			{
				numcol = 16;
				format = Format_INDEXED_16;
			}
			else if (opt.format == "indexed256")
			{
				numcol = 256;
				format = Format_INDEXED_256;
			}
			else if (opt.format == "indexed32a8")
			{
				numcol = 32;
				format = Format_INDEXED_32A8;
			}
			else if (opt.format == "indexed8a32")
			{
				numcol = 8;
				format = Format_INDEXED_8A32;
			}

			Palette palette;
			std::vector<uint16_t> palette16;
			std::vector<Image> images(opt.path_input.size());
			
			for (int i=0; i<images.size(); i++)
				images[i].Load(opt.path_input[i]);

			auto remap = Quantize(images, palette, numcol, false, format);
			
			palette16.resize(palette.size());
			for (int i=1; i<palette.size(); i++)
				palette16[i] = RGB255_RGB16(palette[i].r, palette[i].g, palette[i].b, palette[i].a);

			uint32_t hash = murmurhash(reinterpret_cast<const char*>(palette16.data()), sizeof(uint16_t)*palette16.size(), 0);

			for (int i=0; i<images.size(); i++)
			{
				Metadata meta;
				meta.format = format;
				meta.width = images[i].width;
				meta.height = images[i].height;
				meta.palette_count = palette16.size();
				meta.palette_hash = hash;
				meta.original_size = remap[i].size();

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(2, remap[i].data(), meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(opt.path_input[i]).stem().string());
				std::ofstream out(output, std::ios::binary);
				
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(palette16.data()), sizeof(uint16_t)*palette.size());
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}
	}
	catch(const std::exception &e)
	{
		std::cerr << std::format("Exception caught: {}\n", e.what());
		return 1;
	}
	catch (...)
	{
		std::cerr << std::format("An unknown exception occurred\n");
		return 1;
	}

	std::cout  << "done\n";

	return 0;
}

// // ----

void Image::Load(const std::filesystem::path &path)
{
	int comp;
	stbi_uc *ptr_img = stbi_load(path.c_str(), &width, &height, &comp, 4);
	if (ptr_img == nullptr) throw std::runtime_error(stbi_failure_reason());

	int size = width*height*comp;
	data.assign(ptr_img, ptr_img+size);

	stbi_image_free(ptr_img);
}