#include <libimagequant.h>
#include <stdexcept>
#include "header.hpp"

std::vector<ImageMapped> Quantize(const std::vector<Image>& images, Palette &palette_output, int num_colors, bool dither, uint8_t format)
{
	liq_attr *attr = liq_attr_create();
	liq_histogram *hist = liq_histogram_create(attr);

	liq_set_max_colors(attr, num_colors);

	std::vector<liq_image*> liq_images;
	liq_images.reserve(images.size());

	for (const Image &img : images)
	{
		liq_image *liq_img = liq_image_create_rgba(attr, img.data.data(), img.width, img.height, 0);
		liq_histogram_add_image(hist, attr, liq_img);
		liq_images.push_back(liq_img);
	}

	liq_result *result;
	if (LIQ_OK != liq_histogram_quantize(hist, attr, &result))
		throw std::runtime_error("failed to quantize histogram");

	if (dither) liq_set_dithering_level(result, 1.0f);
	else liq_set_dithering_level(result, 0.0f);

	const liq_palette *pal = liq_get_palette(result);
	palette_output.resize(pal->count);
	for (int i=0; i<pal->count; i++)
	{
		palette_output[i].r = pal->entries[i].r;
		palette_output[i].g = pal->entries[i].g;
		palette_output[i].b = pal->entries[i].b;
		palette_output[i].a = pal->entries[i].a;
	}

	std::vector<ImageMapped> output(images.size());

	for (int i=0; i<images.size(); i++)
	{
		output[i].resize(images[i].width*images[i].height);
		if (LIQ_OK != liq_write_remapped_image(result, liq_images[i], output[i].data(), output[i].size()))
			throw std::runtime_error("failed to remap image");
	}

	for (auto &i : liq_images) liq_image_destroy(i);
	liq_histogram_destroy(hist);
	liq_attr_destroy(attr);

	// ---

	if (format == Format_INDEXED_4)
	{
		std::vector<ImageMapped> packed(output.size());

		for (int i=0; i<output.size(); i++)
		{
			packed[i].reserve((output[i].size()+3) / 4);

			for (int chunk_start=0; chunk_start<output[i].size(); chunk_start+=4) {
				uint8_t current_byte = 0;

				for (int offset=0; offset<4 && (chunk_start+offset)<output[i].size(); ++offset) {
					current_byte |= (output[i][chunk_start + offset] & 0b11) << (offset * 2);
				}

				packed[i].push_back(current_byte);
			}
		}
		
		return packed;
	}

	else if (format == Format_INDEXED_16)
	{
		std::vector<ImageMapped> packed(output.size());

		for (int i=0; i<output.size(); i++)
		{
			packed[i].reserve((output[i].size()+3) / 4);

			for (int offset=0; offset<output[i].size(); offset+=2) {
				uint8_t current_byte = 0;

				current_byte |= (output[i][offset] & 0b1111);
				if (i+1 <output[i].size())
					current_byte |= (output[i][offset + 1] & 0b1111) << 4;

				packed[i].push_back(current_byte);
			}
		}

		return packed;
	}

	else if (format == Format_INDEXED_8A32)
	{
		std::vector<ImageMapped> packed(output.size());

		for (int i=0; i<output.size(); i++)
		{
			packed[i].reserve((output[i].size()+3) / 4);

			for (int offset=0; offset<output[i].size(); offset++) {
				uint8_t current_byte = 0;

				uint8_t alpha = images[i].data[offset*4+3] >> 3;
				current_byte |= (output[i][offset] & 0b111) | (alpha << 3);
				packed[i].push_back(current_byte);
			}
		}

		return packed;
	}

	else if (format == Format_INDEXED_32A8)
	{
		std::vector<ImageMapped> packed(output.size());

		for (int i=0; i<output.size(); i++)
		{
			packed[i].reserve((output[i].size()+3) / 4);

			for (int offset=0; offset<output[i].size(); offset++) {
				uint8_t current_byte = 0;

				uint8_t alpha = images[i].data[offset*4+3] >> 5;
				current_byte |= (output[i][offset] & 0b11111) | (alpha << 5);
				packed[i].push_back(current_byte);
			}
		}

		return packed;
	}

	return output;
}